@@ -35,6 +35,24 @@
 
 #include <glib/gi18n-lib.h>
 
+/**
+ * POS_INPUT_SURFACE_IS_LANG_LAYOUT:
+ * @layout: The layout to check
+ *
+ * Is this widget a "regular" language layout (not emoji, not terminal, …)?
+ */
+#define POS_INPUT_SURFACE_IS_LANG_LAYOUT(widget) \
+  (POS_IS_OSK_WIDGET ((widget)) && GTK_WIDGET ((widget)) != self->osk_terminal)
+
+/**
+ * POS_INPUT_SURFACE_IS_TERMINAL_LAYOUT:
+ * @layout: The layout to check
+ *
+ * Is this widget a terminal layout?
+ */
+#define POS_INPUT_SURFACE_IS_TERMINAL_LAYOUT(widget) \
+  (POS_IS_OSK_WIDGET ((widget)) && GTK_WIDGET ((widget)) == self->osk_terminal)
+
 enum {
   PROP_0,
   PROP_INPUT_METHOD,
@@ -54,6 +72,7 @@ typedef struct {
   gboolean show;
   double   progress;
   gint64   last_frame;
+  guint    id;
 } PosInputSurfaceAnimation;
 
 /**
@@ -148,7 +167,7 @@ pos_input_surface_toggle_shortcuts_bar (PosInputSurface *self)
   child = hdy_deck_get_visible_child (self->deck);
 
   /* shortcuts bar is only for terminal and when we have shortcuts defined */
-  if (child == self->osk_terminal)
+  if (POS_INPUT_SURFACE_IS_TERMINAL_LAYOUT (child))
     shortcuts_visible = !!pos_shortcuts_bar_get_num_shortcuts (self->shortcuts_bar);
 
   gtk_widget_set_visible (GTK_WIDGET (self->shortcuts_bar), shortcuts_visible);
@@ -370,6 +389,44 @@ on_osk_key_symbol (PosInputSurface *self, const char *symbol, GtkWidget *osk_wid
 }
 
 
+static void
+set_keymap (PosInputSurface *self)
+{
+  GtkWidget *child;
+  PosOskWidget *osk;
+
+  /* If the object isn't fully constructed */
+  if (self->keyboard_driver == NULL)
+    return;
+
+  child = hdy_deck_get_visible_child (self->deck);
+  /* Do nothing on e.g. debug surface */
+  if (!POS_IS_OSK_WIDGET (child))
+    return;
+
+  osk = POS_OSK_WIDGET (child);
+  if (osk == POS_OSK_WIDGET (self->osk_terminal)) {
+    pos_vk_driver_set_terminal_keymap (self->keyboard_driver);
+  } else {
+    pos_vk_driver_set_keymap_symbols (self->keyboard_driver,
+                                      pos_osk_widget_get_layout_id (osk),
+                                      pos_osk_widget_get_symbols (osk));
+  }
+}
+
+
+static void
+set_keymap_delayed (PosInputSurface *self)
+{
+  /*
+   * Add a slight delay before switching back the keymap. Otherwise an
+   * X11 client might apply the symbol sent to the popup to the new
+   * keymap.
+   */
+  g_timeout_add_once (25, (GSourceOnceFunc)set_keymap, self);
+}
+
+
 static void
 on_osk_mode_changed (PosInputSurface *self, GParamSpec *pspec, GtkWidget *osk_widget)
 {
@@ -384,6 +441,25 @@ on_osk_mode_changed (PosInputSurface *self, GParamSpec *pspec, GtkWidget *osk_wi
 }
 
 
+static void
+on_osk_popover_shown (PosInputSurface *self, GStrv *symbols, GtkWidget *osk_widget)
+{
+  g_return_if_fail (POS_IS_INPUT_SURFACE (self));
+  g_return_if_fail (POS_IS_OSK_WIDGET (osk_widget));
+
+  pos_vk_driver_set_overlay_keymap (self->keyboard_driver, (const char * const *)symbols);
+}
+
+
+static void
+on_osk_popover_hidden (PosInputSurface *self)
+{
+  g_return_if_fail (POS_IS_INPUT_SURFACE (self));
+
+  set_keymap_delayed (self);
+}
+
+
 static void
 clipboard_copy_activated (GSimpleAction *action,
                           GVariant      *parameter,
@@ -409,14 +485,51 @@ clipboard_paste_activated (GSimpleAction *action,
 
 /* Emoji picker */
 
+
+static void
+send_emoji_via_vk (PosInputSurface *self, const char *emoji)
+{
+  g_autoptr (GPtrArray) syms_array = g_ptr_array_new_with_free_func (g_free);
+  g_autofree gunichar *items = NULL;
+
+  /* To send an emoji via the vk driver we split it into unicode symbols… */
+  items = g_utf8_to_ucs4_fast (emoji, -1, NULL);
+  for (int i = 0; items[i]; i++) {
+    char *symbol = g_new0 (char, 7);
+
+    g_unichar_to_utf8 (items[i], symbol);
+    /* TODO: Can use g_strv_builder_take with glib 2.80 */
+    g_ptr_array_add (syms_array, symbol);
+  }
+  g_ptr_array_add (syms_array, NULL);
+
+  /* …add a keymap that contains the emoji, combining characters and emoji modifiers */
+  pos_vk_driver_set_overlay_keymap (self->keyboard_driver, (const char * const*)syms_array->pdata);
+
+  /* … and type each of these symbols one by one */
+  for (int i = 0; syms_array->pdata[i]; i++) {
+    const char *symbol = syms_array->pdata[i];
+
+    pos_vk_driver_key_down (self->keyboard_driver, symbol);
+    pos_vk_driver_key_up (self->keyboard_driver, symbol);
+  }
+
+  set_keymap_delayed (self);
+}
+
+
 static void
 on_emoji_picked (PosInputSurface *self, const char *emoji, PosEmojiPicker *emoji_picker)
 {
   g_assert (POS_IS_INPUT_SURFACE (self));
   g_assert (POS_IS_EMOJI_PICKER (emoji_picker));
 
-  pos_input_surface_submit_current_preedit (self);
-  pos_input_method_send_string (self->input_method, emoji, TRUE);
+  if (pos_input_method_get_active (self->input_method)) {
+    pos_input_surface_submit_current_preedit (self);
+    pos_input_method_send_string (self->input_method, emoji, TRUE);
+  } else {
+    send_emoji_via_vk (self, emoji);
+  }
 
   pos_input_surface_notify_key_press (self);
 }
@@ -438,6 +551,19 @@ on_emoji_picker_delete_last (PosInputSurface *self)
 
 /* menu button */
 
+
+static const char *
+pos_osk_get_display_name (PosOskWidget *osk_widget)
+{
+  PosCompletionInfo *info = g_object_get_data (G_OBJECT (osk_widget), "pos-completion-info");
+
+  if (info)
+    return info->display_name;
+
+  return pos_osk_widget_get_display_name (osk_widget);
+}
+
+
 static void
 menu_add_layout (gpointer key, gpointer value, gpointer data)
 {
@@ -445,7 +571,7 @@ menu_add_layout (gpointer key, gpointer value, gpointer data)
   const char *name = key;
   PosOskWidget *osk_widget = POS_OSK_WIDGET (value);
   PosInputSurface *self = POS_INPUT_SURFACE (data);
-  const char *display_name = pos_osk_widget_get_display_name (osk_widget);
+  const char *display_name = pos_osk_get_display_name (osk_widget);
 
   button = g_object_new (GTK_TYPE_MODEL_BUTTON,
                          "visible", TRUE,
@@ -487,8 +613,7 @@ menu_activated (GSimpleAction *action, GVariant *parameter, gpointer data)
   g_debug ("Menu popoup activated at %d %d, current: '%s'", rect.x, rect.y, osk_name);
 
   layout_action = g_action_map_lookup_action (G_ACTION_MAP (self->action_map), "select-layout");
-  g_simple_action_set_state (G_SIMPLE_ACTION (layout_action),
-                             g_variant_new ("s", osk_name));
+  g_simple_action_set_state (G_SIMPLE_ACTION (layout_action), g_variant_new ("s", osk_name));
 
   gtk_container_foreach (GTK_CONTAINER (self->menu_box_layouts),
                          (GtkCallback) gtk_widget_destroy,
@@ -543,25 +668,46 @@ select_layout_change_state (GSimpleAction *action,
 }
 
 
+static void pos_input_surface_set_completer (PosInputSurface *self, PosCompleter *completer);
+
+/* Switch the completion engine and it's configuration */
 static void
-switch_language (PosInputSurface *self, const char *locale, const char *region)
+pos_input_surface_switch_completion (PosInputSurface *self, PosOskWidget *osk)
 {
+  PosCompletionInfo *info;
   gboolean success;
+  PosCompleter *default_completer;
   g_autoptr (GError) err = NULL;
 
-  if (self->completer == NULL)
-    return;
+  default_completer = pos_completer_manager_get_default_completer (self->completer_manager);
 
-  g_debug ("Switching language, locale: '%s-%s'", locale, region);
-  success = pos_completer_set_language (self->completer, locale, region, &err);
-  if (success == FALSE) {
-    g_warning ("Failed to set language: %s-%s: %s, switching to '%s-%s' instead",
-               locale, region, err->message, POS_COMPLETER_DEFAULT_LANG,
-               POS_COMPLETER_DEFAULT_REGION);
-    pos_completer_set_language (self->completer,
-                                POS_COMPLETER_DEFAULT_LANG,
-                                POS_COMPLETER_DEFAULT_REGION,
-                                NULL);
+  info = g_object_get_data (G_OBJECT (osk), "pos-completion-info");
+  if (info) {
+    /* Layout with completion info */
+    pos_input_surface_set_completer (self, info->completer);
+    success = pos_completer_set_language (self->completer, info->lang, info->region, &err);
+    if (!success)
+      g_warning ("Failed to switch completer: %s", err->message);
+  } else if (default_completer) {
+    /* Layout without completion info - use default completer */
+    const char *lang = pos_osk_widget_get_lang (osk);
+    const char *region = pos_osk_widget_get_region (osk);
+
+    pos_input_surface_set_completer (self, default_completer);
+    success = pos_completer_set_language (self->completer, lang, region, &err);
+    if (!success) {
+      g_warning ("Failed to set completion language: %s-%s: %s, switching to '%s-%s' instead",
+                 lang, region, err->message, POS_COMPLETER_DEFAULT_LANG,
+                 POS_COMPLETER_DEFAULT_REGION);
+      g_clear_error (&err);
+      if (!pos_completer_set_language (self->completer,
+                                       POS_COMPLETER_DEFAULT_LANG,
+                                       POS_COMPLETER_DEFAULT_REGION,
+                                       &err)) {
+        g_warning ("Failed to set completion language '%s-%s': %s",
+                   POS_COMPLETER_DEFAULT_LANG, POS_COMPLETER_DEFAULT_REGION, err->message);
+      }
+    }
   }
 
   pos_completion_bar_set_completions (POS_COMPLETION_BAR (self->completion_bar), NULL);
@@ -585,11 +731,13 @@ on_visible_child_changed (PosInputSurface *self)
   g_debug ("Switched to layout '%s'", pos_osk_widget_get_display_name (osk));
   pos_osk_widget_set_layer (osk, POS_OSK_WIDGET_LAYER_NORMAL);
 
-  /* Remember last layout */
-  self->last_layout = GTK_WIDGET (osk);
+  set_keymap (self);
 
-  if (osk != POS_OSK_WIDGET (self->osk_terminal))
-    switch_language (self, pos_osk_widget_get_lang (osk), pos_osk_widget_get_region (osk));
+  /* Remember last layout */
+  if (POS_INPUT_SURFACE_IS_LANG_LAYOUT (osk)) {
+    pos_input_surface_switch_completion (self, osk);
+    self->last_layout = GTK_WIDGET (osk);
+  }
 
   /* Recheck completion bar visibility */
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_COMPLETER_ACTIVE]);
@@ -677,7 +825,6 @@ animate_cb (GtkWidget     *widget,
   return G_SOURCE_CONTINUE;
 }
 
-
 static void
 pos_input_surface_set_completer (PosInputSurface *self, PosCompleter *completer)
 {
@@ -701,9 +848,6 @@ pos_input_surface_set_completer (PosInputSurface *self, PosCompleter *completer)
                       "swapped-signal::update",
                       G_CALLBACK (on_completer_update), self,
                       NULL);
-    switch_language (self,
-                     pos_osk_widget_get_lang (POS_OSK_WIDGET (self->last_layout)),
-                     pos_osk_widget_get_region (POS_OSK_WIDGET (self->last_layout)));
   } else {
     g_debug ("Removing completer");
   }
@@ -716,16 +860,14 @@ static void
 pos_input_surface_set_completer_manager (PosInputSurface     *self,
                                          PosCompleterManager *completer_manager)
 {
-  PosCompleter *default_completer;
-
   if (self->completer_manager == completer_manager)
     return;
 
   g_set_object (&self->completer_manager, completer_manager);
 
-  /* Update the current completer */
-  default_completer = pos_completer_manager_get_default_completer (self->completer_manager);
-  pos_input_surface_set_completer (self, default_completer);
+  /* Switch completion */
+  if (self->last_layout)
+    pos_input_surface_switch_completion (self, POS_OSK_WIDGET (self->last_layout));
 }
 
 
@@ -855,6 +997,7 @@ on_im_purpose_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod
 {
   GtkWidget *osk_widget = NULL;
   PosOskWidgetLayer layer = POS_OSK_WIDGET_LAYER_NORMAL;
+  PosInputMethodPurpose purpose;
 
   g_assert (POS_IS_INPUT_SURFACE (self));
   g_assert (POS_IS_INPUT_METHOD (im));
@@ -862,7 +1005,8 @@ on_im_purpose_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod
   /* We only have completer active on `normal` input purpose */
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_COMPLETER_ACTIVE]);
 
-  switch (pos_input_method_get_purpose (im)) {
+  purpose = pos_input_method_get_purpose (im);
+  switch (purpose) {
   case POS_INPUT_METHOD_PURPOSE_ALPHA:
   case POS_INPUT_METHOD_PURPOSE_EMAIL:
   case POS_INPUT_METHOD_PURPOSE_NAME:
@@ -889,10 +1033,13 @@ on_im_purpose_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod
 
   if (osk_widget == NULL) {
     osk_widget = hdy_deck_get_visible_child (self->deck);
-    /* Debug surface and emoji don't have layers */
-    if (POS_IS_OSK_WIDGET (osk_widget) == FALSE)
+    /* If no "special" layout, Switch back to the last language layer */
+    if (!POS_INPUT_SURFACE_IS_LANG_LAYOUT (osk_widget))
       osk_widget = self->last_layout;
   }
+  g_debug ("Layout: %s, purpose: %d",
+           pos_osk_widget_get_name (POS_OSK_WIDGET (osk_widget)),
+           purpose);
   hdy_deck_set_visible_child (self->deck, osk_widget);
 
   pos_osk_widget_set_layer (POS_OSK_WIDGET (osk_widget), layer);
@@ -976,14 +1123,30 @@ on_im_active_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod *
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_COMPLETER_ACTIVE]);
 }
 
+static PosOskWidget *insert_xkb_layout (PosInputSurface *self, const char *type, const char *id);
+static void on_input_setting_changed (PosInputSurface *self, const char *key, GSettings *settings);
 
 static void
 pos_input_surface_constructed (GObject *object)
 {
   PosInputSurface *self = POS_INPUT_SURFACE (object);
+  const char *test_layout = g_getenv ("POS_TEST_LAYOUT");
 
   G_OBJECT_CLASS (pos_input_surface_parent_class)->constructed (object);
 
+  if (test_layout) {
+    PosOskWidget *osk_widget = insert_xkb_layout (self, "xkb", test_layout);
+    hdy_deck_set_visible_child (self->deck, GTK_WIDGET (osk_widget));
+  } else {
+    g_object_connect (self->input_settings,
+                      "swapped-signal::changed::sources",
+                      G_CALLBACK (on_input_setting_changed), self,
+                      "swapped-signal::changed::xkb-options",
+                      G_CALLBACK (on_input_setting_changed), self,
+                      NULL);
+    on_input_setting_changed (self, NULL, self->input_settings);
+  }
+
   g_assert (POS_IS_INPUT_METHOD (self->input_method));
   g_object_connect (self->input_method,
                     "swapped-object-signal::notify::active", on_im_active_changed, self,
@@ -994,6 +1157,8 @@ pos_input_surface_constructed (GObject *object)
                     "swapped-object-signal::notify::surrounding-text",
                     on_im_surrounding_text_changed, self,
                     NULL);
+
+  set_keymap (self);
 }
 
 
@@ -1018,6 +1183,8 @@ pos_input_surface_finalize (GObject *object)
   g_signal_remove_emission_hook (g_signal_lookup ("clicked", GTK_TYPE_BUTTON), self->clicked_id);
   self->clicked_id = 0;
 
+  g_clear_handle_id (&self->animation.id, g_source_remove);
+
   g_clear_object (&self->input_method);
   g_clear_object (&self->a11y_settings);
   g_clear_object (&self->input_settings);
@@ -1213,6 +1380,8 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, on_osk_key_down);
   gtk_widget_class_bind_template_callback (widget_class, on_osk_key_symbol);
   gtk_widget_class_bind_template_callback (widget_class, on_osk_mode_changed);
+  gtk_widget_class_bind_template_callback (widget_class, on_osk_popover_shown);
+  gtk_widget_class_bind_template_callback (widget_class, on_osk_popover_hidden);
   gtk_widget_class_bind_template_callback (widget_class, on_shortcut_activated);
   gtk_widget_class_bind_template_callback (widget_class, on_visible_child_changed);
 
@@ -1226,12 +1395,12 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
                          POS_TYPE_INPUT_METHOD,
                          G_PARAM_READWRITE |
                          G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
-   /**
-    * PosInputSurface:completer:
-    *
-    * A completer implementing the #PosCompleter interface.
-    *
-    */
+  /**
+   * PosInputSurface:completer:
+   *
+   * A completer implementing the #PosCompleter interface.
+   *
+   */
    props[PROP_COMPLETER] =
      g_param_spec_object ("completer", "", "",
                           POS_TYPE_COMPLETER,
@@ -1276,14 +1445,18 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
   /**
    * PosInputSurface:completion-enabled
    *
-   * %TRUE if the user enabled completion. That does imply that completion is actually active
+   * %TRUE if the user enabled completion. That does not imply that completion is actually active
    * as this also depends on an input-method being present, a completer configured, etc. This
    * setting merely reflects the users intent.
    */
   props[PROP_COMPLETION_ENABLED] =
     g_param_spec_boolean ("completion-enabled", "", "", FALSE,
                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
-
+  /**
+   * PosInputSurface:keyboard-driver:
+   *
+   * Keyboard driver for submitting keycodes and handling associated keymaps.
+   */
   props[PROP_KEYBOARD_DRIVER] =
     g_param_spec_object ("keyboard-driver", "", "",
                          /* TODO: should be an interface */
@@ -1304,12 +1477,28 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
 }
 
 
+static char *
+build_layout_name (const char *engine, const char *layout, const char *variant)
+{
+  char *name;
+
+  if (STR_IS_NULL_OR_EMPTY (variant))
+    name = g_strdup_printf ("%s:%s", engine, layout);
+  else
+    name = g_strdup_printf ("%s:%s+%s", engine, layout, variant);
+
+  return name;
+}
+
+
 static PosOskWidget *
-insert_osk (PosInputSurface *self,
-            const char      *name,
-            const char      *display_name,
-            const char      *layout,
-            const char      *variant)
+insert_osk (PosInputSurface   *self,
+            const char        *name,
+            const char        *layout_id,
+            const char        *display_name,
+            const char        *layout,
+            const char        *variant,
+            PosCompletionInfo *info)
 {
   g_autoptr (GError) err = NULL;
   PosOskWidget *osk_widget;
@@ -1320,25 +1509,37 @@ insert_osk (PosInputSurface *self,
 
   osk_widget = pos_osk_widget_new (self->osk_features);
   if (!pos_osk_widget_set_layout (POS_OSK_WIDGET (osk_widget),
-                                  display_name, layout, variant, &err)) {
+                                  name,
+                                  layout_id,
+                                  display_name,
+                                  layout,
+                                  variant,
+                                  &err)) {
     g_warning ("Failed to load osk layout for %s: %s", name, err->message);
 
     gtk_widget_destroy (g_object_ref_sink (GTK_WIDGET (osk_widget)));
     return NULL;
   }
 
+  g_object_set_data_full (G_OBJECT (osk_widget),
+                          "pos-completion-info",
+                          info,
+                          (GDestroyNotify)pos_completion_info_free);
+
   g_debug ("Adding osk for layout '%s'", name);
   gtk_widget_set_visible (GTK_WIDGET (osk_widget), TRUE);
   g_object_connect (osk_widget,
                     "swapped-signal::key-down", G_CALLBACK (on_osk_key_down), self,
                     "swapped-signal::key-symbol", G_CALLBACK (on_osk_key_symbol), self,
                     "swapped-signal::notify::mode", G_CALLBACK (on_osk_mode_changed), self,
+                    "swapped-signal::popover-shown", G_CALLBACK (on_osk_popover_shown), self,
+                    "swapped-signal::popover-hidden", G_CALLBACK (on_osk_popover_hidden), self,
                     NULL);
 
   hdy_deck_insert_child_after (self->deck, GTK_WIDGET (osk_widget), NULL);
   g_hash_table_insert (self->osks, g_strdup (name), osk_widget);
 
-  if (self->last_layout == NULL)
+  if (self->last_layout == NULL && POS_INPUT_SURFACE_IS_LANG_LAYOUT (osk_widget))
     self->last_layout = GTK_WIDGET (osk_widget);
 
   return osk_widget;
@@ -1346,7 +1547,7 @@ insert_osk (PosInputSurface *self,
 
 
 static PosOskWidget *
-insert_layout (PosInputSurface *self, const char *type, const char *id)
+insert_xkb_layout (PosInputSurface *self, const char *type, const char *layout_id)
 {
   g_autofree char *name = NULL;
   const gchar *layout = NULL;
@@ -1354,21 +1555,57 @@ insert_layout (PosInputSurface *self, const char *type, const char *id)
   const gchar *display_name = NULL;
 
   if (g_strcmp0 (type, "xkb")) {
-    g_debug ("Not a xkb layout: '%s' - ignoring", id);
+    g_debug ("Not a xkb layout: '%s' - ignoring", layout_id);
     return NULL;
   }
 
-  if (!gnome_xkb_info_get_layout_info (self->xkbinfo, id, &display_name, NULL,
+  if (!gnome_xkb_info_get_layout_info (self->xkbinfo, layout_id, &display_name, NULL,
                                        &layout, &variant)) {
-    g_warning ("Failed to get layout info for %s", id);
+    g_warning ("Failed to get layout info for %s", layout_id);
     return NULL;
   }
-  if (STR_IS_NULL_OR_EMPTY (variant))
-    name = g_strdup (layout);
-  else
-    name = g_strdup_printf ("%s+%s", layout, variant);
+  name = build_layout_name ("xkb", layout, variant);
 
-  return insert_osk (self, name, display_name, layout, variant);
+  return insert_osk (self, name, layout_id, display_name, layout, variant, NULL);
+}
+
+static PosOskWidget *
+insert_ibus_layout (PosInputSurface *self, const char *type, const char *id)
+{
+  const char *engine_name, *lang;
+  g_autofree char *name = NULL;
+  g_autoptr (GError) err = NULL;
+  g_auto (GStrv) parts = NULL;
+  PosCompletionInfo *info;
+
+  /* We don't actually do ibus bus but try to match these to completers */
+  if (g_strcmp0 (type, "ibus")) {
+    g_debug ("Not an ibus layout: '%s' - ignoring", id);
+    return NULL;
+  }
+
+  parts = g_strsplit (id, ":", -1);
+  if (g_strv_length (parts) > 3) {
+    g_warning ("ibus layout '%s' not parsable - ignoring", id);
+    return NULL;
+  }
+  engine_name = parts[0];
+  lang = parts[1];
+
+  info = pos_completer_manager_get_info (self->completer_manager, engine_name, lang, NULL, &err);
+  if (!info) {
+    g_warning ("ibus layout '%s': engine '%s' not usable for '%s': %s - ignoring",
+               id,
+               engine_name,
+               lang,
+               err->message);
+    return NULL;
+  }
+
+  name = build_layout_name ("ibus", lang, NULL);
+
+  /* TODO: allow for other base layouts than "us" */
+  return insert_osk (self, name, id, info->display_name, "us", NULL, info);
 }
 
 
@@ -1398,7 +1635,10 @@ on_input_setting_changed (PosInputSurface *self, const char *key, GSettings *set
   while (g_variant_iter_next (&iter, "(&s&s)", &type, &id)) {
     PosOskWidget *osk_widget;
 
-    osk_widget = insert_layout (self, type, id);
+    osk_widget = insert_xkb_layout (self, type, id);
+    if (osk_widget == NULL)
+      osk_widget = insert_ibus_layout (self, type, id);
+
     if (osk_widget == NULL)
       continue;
 
@@ -1421,8 +1661,10 @@ on_input_setting_changed (PosInputSurface *self, const char *key, GSettings *set
 
   /* If nothing is left add a default */
   if (g_hash_table_size (self->osks) == 0) {
-    insert_osk (self, "us", "English (USA)", "us", NULL);
+    insert_osk (self, "us", "us", "English (USA)", "us", NULL, NULL);
   }
+
+  set_keymap (self);
 }
 
 
@@ -1470,8 +1712,8 @@ static void
 pos_input_surface_init (PosInputSurface *self)
 {
   GtkSettings *gtk_settings;
-  const char *test_layout = g_getenv ("POS_TEST_LAYOUT");
   g_autoptr (GPropertyAction) completion_action = NULL;
+  g_autoptr (GError) err = NULL;
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
@@ -1506,22 +1748,14 @@ pos_input_surface_init (PosInputSurface *self)
   on_completion_mode_changed (self, NULL, self->osk_settings);
   g_settings_bind (self->osk_settings, "osk-features", self, "osk-features", G_SETTINGS_BIND_GET);
 
-  pos_osk_widget_set_layout (POS_OSK_WIDGET (self->osk_terminal),
-                             _("Terminal"),
-                             "terminal",
-                             NULL,
-                             NULL);
-  if (test_layout) {
-    PosOskWidget *osk_widget = insert_layout (self, "xkb", test_layout);
-    hdy_deck_set_visible_child (self->deck, GTK_WIDGET (osk_widget));
-  } else {
-    g_object_connect (self->input_settings,
-                      "swapped-signal::changed::sources",
-                      G_CALLBACK (on_input_setting_changed), self,
-                      "swapped-signal::changed::xkb-options",
-                      G_CALLBACK (on_input_setting_changed), self,
-                      NULL);
-    on_input_setting_changed (self, NULL, self->input_settings);
+  if (!pos_osk_widget_set_layout (POS_OSK_WIDGET (self->osk_terminal),
+                                  "terminal",
+                                  "terminal",
+                                  _("Terminal"),
+                                  "terminal",
+                                  NULL,
+                                  &err)) {
+    g_warning ("Failed to set terminal layout: %s", err->message);
   }
 
   gtk_settings = gtk_settings_get_default ();
@@ -1548,11 +1782,28 @@ pos_input_surface_get_active (PosInputSurface *self)
 }
 
 
+static gboolean
+animation_timeout_cb (gpointer data)
+{
+  PosInputSurface *self = POS_INPUT_SURFACE (data);
+
+  if (self->animation.progress < 1.0) {
+    g_warning ("Animation did not finish in time: %f", self->animation.progress);
+    self->animation.progress = 1.0;
+    pos_input_surface_move (self);
+  }
+
+  self->animation.id = 0;
+  return FALSE;
+}
+
+
 void
 pos_input_surface_set_visible (PosInputSurface *self, gboolean visible)
 {
   g_return_if_fail (POS_IS_INPUT_SURFACE (self));
 
+  g_debug ("Showing keyboard: %d, %d", visible, self->surface_visible);
   if (visible == self->surface_visible)
     return;
 
@@ -1564,6 +1815,10 @@ pos_input_surface_set_visible (PosInputSurface *self, gboolean visible)
   self->animation.progress =
     reverse_ease_out_cubic (1.0 - hdy_ease_out_cubic (self->animation.progress));
 
+  if (self->animation.id)
+    g_source_remove (self->animation.id);
+
+  self->animation.id = g_timeout_add_seconds (1, animation_timeout_cb, self);
   gtk_widget_add_tick_callback (GTK_WIDGET (self), animate_cb, NULL, NULL);
 }
 
@@ -1598,12 +1853,18 @@ pos_input_surface_is_completer_active (PosInputSurface *self)
   if (pos_input_method_get_active (self->input_method) == FALSE)
     return FALSE;
 
+  /* Layout has an "implicit" completer (e.g. varnam) */
+  child = hdy_deck_get_visible_child (self->deck);
+  if (POS_IS_OSK_WIDGET (child) &&
+      g_object_get_data (G_OBJECT (child), "pos-completion-info")) {
+      return TRUE;
+  }
+
   if (self->completion_enabled == FALSE)
     return FALSE;
 
   /* Completion should only be used on "regular" layouts */
-  child = hdy_deck_get_visible_child (self->deck);
-  if (POS_IS_OSK_WIDGET (child) == FALSE || child == self->osk_terminal)
+  if (!POS_INPUT_SURFACE_IS_LANG_LAYOUT (child))
     return FALSE;
 
   /* We only complete input purpose `normal` */
