@@ -15,6 +15,7 @@
 #include "pos-enum-types.h"
 #include "pos-osk-key.h"
 #include "pos-osk-widget.h"
+#include "pos-virtual-keyboard.h"
 
 #include <json-glib/json-glib.h>
 #include <pango/pangocairo.h>
@@ -38,6 +39,8 @@ enum {
   OSK_KEY_UP,
   OSK_KEY_CANCELLED,
   OSK_KEY_SYMBOL,
+  OSK_POPOVER_SHOWN,
+  OSK_POPOVER_HIDDEN,
   N_SIGNALS
 };
 static guint signals[N_SIGNALS];
@@ -124,11 +127,14 @@ struct _PosOskWidget {
   GtkStyleContext     *key_context;
   PosOskWidgetLayer    layer;
   PosOskWidgetMode     mode;
+  /* Contains pointers to key symbols (keys have ownership) */
+  GPtrArray           *symbols;
 
   char                *name;
   char                *display_name;
   char                *lang;
   char                *region;
+  char                *layout_id;
 
   PosOskKey           *current;
   PosOskKey           *space;
@@ -436,7 +442,8 @@ parse_row (PosOskWidget *self, PosOskWidgetRow *row, JsonArray *arow, PosOskWidg
     }
 
     row->width += pos_osk_key_get_width (key);
-    g_ptr_array_insert (row->keys, -1, g_steal_pointer (&key));
+    g_ptr_array_add (self->symbols, (gpointer)pos_osk_key_get_symbol (key));
+    g_ptr_array_add (row->keys, g_steal_pointer (&key));
   }
 
   add_common_keys_pre (self, row, l, r);
@@ -613,6 +620,8 @@ parse_layout (PosOskWidget *self, const char *json, gsize size)
   }
   parse_layers (self, levels);
 
+  g_ptr_array_add (self->symbols, NULL);
+
   return TRUE;
 }
 
@@ -963,6 +972,14 @@ on_symbol_selected (PosOskWidget *self, const char *symbol)
 }
 
 
+static void
+on_popover_closed (PosOskWidget *self)
+{
+  g_debug ("Closed symbol popover");
+  g_signal_emit (self, signals[OSK_POPOVER_HIDDEN], 0);
+}
+
+
 static void
 on_long_pressed (GtkGestureLongPress *gesture, double x, double y, gpointer user_data)
 {
@@ -992,8 +1009,16 @@ on_long_pressed (GtkGestureLongPress *gesture, double x, double y, gpointer user
   get_popup_pos (key, &rect);
   gtk_popover_set_pointing_to (GTK_POPOVER (self->char_popup), &rect);
 
-  g_signal_connect_swapped (self->char_popup, "selected", G_CALLBACK (on_symbol_selected), self);
+  g_signal_connect_object (self->char_popup, "selected",
+                           G_CALLBACK (on_symbol_selected),
+                           self,
+                           G_CONNECT_SWAPPED);
+  g_signal_connect_object (self->char_popup, "closed",
+                           G_CALLBACK (on_popover_closed),
+                           self,
+                           G_CONNECT_SWAPPED);
   gtk_popover_popup (GTK_POPOVER (self->char_popup));
+  g_signal_emit (self, signals[OSK_POPOVER_SHOWN], 0, symbols);
 }
 
 
@@ -1277,6 +1302,8 @@ pos_osk_widget_finalize (GObject *object)
   g_clear_pointer (&self->display_name, g_free);
   g_clear_pointer (&self->lang, g_free);
   g_clear_pointer (&self->region, g_free);
+  g_clear_pointer (&self->layout_id, g_free);
+  g_ptr_array_free (self->symbols, FALSE);
 
   G_OBJECT_CLASS (pos_osk_widget_parent_class)->finalize (object);
 }
@@ -1341,7 +1368,9 @@ pos_osk_widget_class_init (PosOskWidgetClass *klass)
   /**
    * PosOskWidget:name
    *
-   * The name of the current layout
+   * The name of the current layout. The name is unique for this layout. For xkb based layouts
+   * it's `xkb:lang:variant`, for special layouts like `terminal` just `terminal`. The widget
+   * should treat this as opaque value.
    */
   props[PROP_NAME] =
     g_param_spec_string ("name", "", "",
@@ -1406,6 +1435,32 @@ pos_osk_widget_class_init (PosOskWidgetClass *klass)
                                           G_TYPE_NONE,
                                           1,
                                           G_TYPE_STRING);
+  /**
+   * PosOskWidget::popover-shown
+   * @self: The osk widget emitting the symbol
+   * @symbols: The symbols in the popover
+   *
+   * The osk shows a popover to select additional symbols
+   */
+  signals[OSK_POPOVER_SHOWN] = g_signal_new ("popover-shown",
+                                             G_TYPE_FROM_CLASS (klass),
+                                             G_SIGNAL_RUN_LAST,
+                                             0, NULL, NULL, NULL,
+                                             G_TYPE_NONE,
+                                             1,
+                                             G_TYPE_STRV);
+  /**
+   * PosOskWidget::popover-hidden
+   * @self: The osk widget emitting the symbol
+   *
+   * The osk has hidden the symbol popover
+   */
+  signals[OSK_POPOVER_HIDDEN] = g_signal_new ("popover-hidden",
+                                              G_TYPE_FROM_CLASS (klass),
+                                              G_SIGNAL_RUN_LAST,
+                                              0, NULL, NULL, NULL,
+                                              G_TYPE_NONE,
+                                              0);
 
   gtk_widget_class_set_css_name (widget_class, "pos-osk-widget");
 }
@@ -1434,13 +1489,13 @@ pos_osk_widget_init (PosOskWidget *self)
 {
   /* TODO support PIN, number, etc */
   const char *purpose_class = "normal";
-
   g_autoptr (GtkWidgetPath) path = NULL;
   GtkStyleContext *key_context;
   GtkStyleContext *context;
 
   self->mode = POS_OSK_WIDGET_MODE_KEYBOARD;
   self->layer = POS_OSK_WIDGET_LAYER_NORMAL;
+  self->symbols = g_ptr_array_new ();
 
   gtk_widget_add_events (GTK_WIDGET (self), GDK_BUTTON_PRESS_MASK |
                          GDK_BUTTON_RELEASE_MASK |
@@ -1539,7 +1594,7 @@ pos_osk_widget_set_layer (PosOskWidget *self, PosOskWidgetLayer layer)
 
 
 static void
-parse_lang (PosOskWidget *self)
+parse_lang (PosOskWidget *self, const char *layout, const char *variant)
 {
   g_auto (GStrv) parts = NULL;
 
@@ -1556,17 +1611,15 @@ parse_lang (PosOskWidget *self)
     return;
   }
 
-  /* Keyboard layout has language (`en`), region is from name (`us`) */
+  /* Keyboard layout has language (`en`), region is from layout (`us`) */
   self->lang = g_strdup (self->layout.locale);
-  if (strchr (self->name, '+') == 0) {
-    self->region = g_strdup (self->name);
+  if (STR_IS_NULL_OR_EMPTY (variant)) {
+    self->region = g_strdup (layout);
     return;
   }
 
-  /* Like above but name also has a layout variant (`in+mal`) */
-  g_strfreev (parts);
-  parts = g_strsplit (self->name, "+", -1);
-  self->region = g_strdup (parts[0]);
+  /* Like above but layout also from variant (`in+mal`) */
+  self->region = g_strdup (variant);
 }
 
 
@@ -1575,10 +1628,15 @@ parse_lang (PosOskWidget *self)
 /**
  * pos_osk_widget_set_layout:
  * @self: The osk widget
+ * @name: The "name" of the layout. This uniquely identifiers the layout. The widget should
+ *  treat this as an opaque value.
+ * @layout_id: The (xkb) layout id. This can differ from the widget layout and variant
+ *  e.g. in the case of terminal where we use a `terminal` layout but an xkb keymap `us`.
+ *  The widget should treat this as opaque value.
  * @display_name: The display name. Should be used when displaying layout information
  *    to the user. (E.g. 'English (US)')
- * @layout: The name of the layout. E.g. `jp`, `de`
- * @variant: The layout variant, e.g. `ch`
+ * @layout: The name of the layout. to set e.g. `jp`, `de`, 'terminal'
+ * @variant: The layout variant to set , e.g. `ch`
  * @err: The error location
  *
  * Sets the widgets keyboard layout.
@@ -1587,44 +1645,48 @@ parse_lang (PosOskWidget *self)
  */
 gboolean
 pos_osk_widget_set_layout (PosOskWidget *self,
+                           const char   *name,
+                           const char   *layout_id,
                            const char   *display_name,
                            const char   *layout,
                            const char   *variant,
                            GError      **err)
 {
-  g_autofree char *name = NULL;
   g_autofree char *path = NULL;
-
   g_autoptr (GBytes) data = NULL;
   const char *json;
   gsize size;
   gboolean ret;
 
-  if (!STR_IS_NULL_OR_EMPTY (variant))
-    name = g_strdup_printf ("%s+%s", layout, variant);
-  else
-    name = g_strdup (layout);
-
   if (g_strcmp0 (self->name, name) == 0)
     return TRUE;
 
   if (self->layout.name)
     pos_osk_widget_layout_free (&self->layout);
   g_free (self->name);
-  self->name = g_steal_pointer (&name);
+  self->name = g_strdup (name);
   g_free (self->display_name);
   self->display_name = g_strdup (display_name);
+  g_free (self->layout_id);
+  self->layout_id = g_strdup (layout_id);
+
+  if (!STR_IS_NULL_OR_EMPTY (variant))
+    path = g_strdup_printf ("/sm/puri/phosh/osk-stub/layouts/%s+%s.json", layout, variant);
+  else
+    path = g_strdup_printf ("/sm/puri/phosh/osk-stub/layouts/%s.json", layout);
 
-  path = g_strdup_printf ("/sm/puri/phosh/osk-stub/layouts/%s.json", self->name);
   data = g_resources_lookup_data (path, 0, err);
   if (data == NULL) {
     return FALSE;
   }
 
+  g_ptr_array_free (self->symbols, TRUE);
+  self->symbols = g_ptr_array_new ();
+
   json = (char*) g_bytes_get_data (data, &size);
   ret = parse_layout (self, json, size);
 
-  parse_lang (self);
+  parse_lang (self, layout, variant);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_NAME]);
 
@@ -1696,7 +1758,7 @@ pos_osk_widget_get_mode (PosOskWidget *self)
  * pos_osk_widget_get_lang:
  * @self: The osk widget
  *
- * Get the language e.g. `en`, `de`.
+ * Get the language e.g. `en`, `de`
  *
  * Returns: The language
  */
@@ -1726,6 +1788,35 @@ pos_osk_widget_get_region (PosOskWidget *self)
   return self->region;
 }
 
+/**
+ * pos_osk_widget_get_layout_id:
+ * @self: The osk widget
+ *
+ * The (xkb) keymap layout_id used with this widget.
+ */
+const char *
+pos_osk_widget_get_layout_id (PosOskWidget *self)
+{
+  g_return_val_if_fail (POS_IS_OSK_WIDGET (self), NULL);
+
+  return self->layout_id;
+}
+
+
+/**
+ * pos_osk_widget_get_symbols:
+ * @self: The osk widget
+ *
+ * Get the symbols on this OSK.
+ */
+const char * const *
+pos_osk_widget_get_symbols (PosOskWidget *self)
+{
+  g_return_val_if_fail (POS_IS_OSK_WIDGET (self), NULL);
+
+  return (const char * const *)self->symbols->pdata;
+}
+
 
 /**
  * pos_osk_widget_set_features:
